'''
    Bakalářská práce
    Scrolling Song Book - zpěvník s automatickým posunem textu
    
    Hana Kalivodová
    4. 5. 2022
    
    Analýza zvukového signálu, určení akordu
'''

from PyQt5.QtCore import QObject, pyqtSignal
import math
import numpy
import scipy.signal
import statistics

# soubor s konstantami
import const

'''
    Funkce vyhodnocující na základě střední hodnoty energií a hodnoty
    příznaku zero crossing, zda nahraný zvuk je ticho, nebo hudba
    Pokud se jedná o ticho, nebude rámec dále zpracováván
    
    @param time_series      rámec zvukového záznamu (pole reálných čísel)
        
    @return True    rámec vyhodnocený jako hudba / hlas
    @return False   rámec vyhodnocený jako ticho / šum
'''
def is_acceptable(time_series):
    pocet = len((numpy.where((numpy.sign(time_series)) == 0))[0])
    for i in range (0, len(time_series) - 1):
        if numpy.sign(time_series)[i] == 0: continue
        if (numpy.sign(time_series)[i] + numpy.sign(time_series)[i+1]) == 0:
            pocet += 1                

    abs_win = numpy.absolute(time_series)
    mean = statistics.mean(abs_win)

    if (mean < 1e-03 or pocet > 4096):
        return False
    
    return True
        

'''
    Funkce provede přenásobení každého vzorku koeficientem dle Hammingovy funkce okna.
    
    @param array    rámec zvukového záznamu (pole reálných čísel)
    
    @return win_array   rámec zvukového záznamu vynásobený Hammingovo funkcí okna
'''
def get_windowing(array):
    i = 0
    win_array = []
    for i in range(len(array)):
        # koeficient Hammingovy funkce okna
        w_i = (25/46) - (21/46) * math.cos(2 * math.pi * i / (const.SAMPLES))
        win_array.append(array[i] * w_i)
    return win_array

'''
    Funkce provede rychlou Fourierovu transformaci (FFT). 
    Výsledek FFT je v komplexním tvaru.
    
    @param array    pole reálných hodnot, rámec
    
    @return fft_array   výsledek FFT, pole komplexních čísel
'''
def get_fft(array):        
    fft_array = numpy.fft.fft(array)
    return fft_array

'''
    Funkce pro získání spektra z rychlé Fourierovy transformace (FFT)
    Funkce provede výpočet velikosti vektoru z komplexního čísla
    FFT je téměř symetrická, stačí tedy zpracovat polovinu dat
    Vznikne odhad výkonové spektrální hustoty poloviční délky vstupního pole
    
    @param array     pole komplexních čísel (výsledek FFT)
    
    @return spectrum    odhad výkonové spektrální hustoty (pole kladných reálných čísel)
'''
def get_spectrum(array):
    spectrum = []
    for x in range(len(array)//2):
        real = array[x + 1].real
        imag = array[x + 1].imag
        size = math.sqrt((real * real) + (imag * imag))
        spectrum.append(size)
    return spectrum

'''
    Funkce provede průměrování spekter 
    (spektrem je myšlený odhad spektrální výkonové hustoty).
    
    @param list_sp  seznam spekter
    
    @return average     průměr spekter
'''    
def average_spectrals(list_sp):
    average = []
   
    for i in range(len(list_sp[0])):
        x = 0
        for j in range(len(list_sp)):
            x += list_sp[j][i]
        x /= len(list_sp)
        average.append(x)
    
    return average   

'''
    Funkce vynásobí odhad výkonové spektrální hustoty (PSD) logaritmem, 
    aby došlo ke zdůraznění nízkých frekvencí a útlumu vysokých frekvencí.
    
    @param array    odhad PSD
    
    @return log     odhad PSD vynásobený logaritmem
'''
def get_logarithm(array):
    log = []
    for i in range(1, len(array)):
        log.append((-math.log(i) + 8.5) * array[i])     
    
    return log    
    
'''
    Funkce pro výpočet frekvencí
    Vstupem pro výpočet je index vrcholu odhadu výkonové spektrální hustoty (PSD), 
    pro který chceme znát frekvenci
    
    @param array  indexy vrcholů PSD
    
    @return freq    frekvence vrcholů PSD
'''            
def get_freqs(array):
    freq = []
    for x in array:
        freq.append(x*(const.SAMPLE_RATE/const.SAMPLES))
        
    return freq    

'''
    Funkce přiřadí frekvenci konkrétní tón, jejichž rozdíl frekvencí je nejmenší.
    
    @param freq     pole frekvencí
    
    @return res     seznam tónů, které odpovídají frekvencím na vstupu
'''    
def get_notes(freq):
    maximum = const.MAX_FREQ
    minimum = const.MIN_FREQ
    
    tones = ['A', 'Ais', 'H', 'C', 'Cis', 'D', 'Dis', 'E', 'F', 'Fis', 'G', 'Gis']
    frequencies = const.FREQUENCIES
               
    res = []
    for f in freq:
        if (f < minimum or f > maximum):
            continue
        
        variances = []
        for x in frequencies:
            variances.append(abs(x - f))
  
        min_var = min(variances)
        index = variances.index(min_var)
        index %= len(tones)
        res.append(tones[index])

    return res
 
'''
    Funkce počítající podmíněnou pravděpodobnost u všech akordů
    Výsledkem jsou první tři akordy s nejvyšší pravděpodobností.
    
    @param notes        seznam tónů (text)
    @param chords       slovník akordů, každý akord obsahuje seznam tónů, z nichž se skládá
                        a počet shodných tónů s tóny ze seznamu notes
    
    @return chord       3 nejpravděpodobnější akordy (akordy s největším počtem shod)
''' 
def get_chord(notes, chords):
    tones = {'C': 0, 'Cis': 0, 'D': 0, 'Dis': 0, 'E': 0, 'F': 0, 'Fis': 0, 'G': 0, 'Gis': 0, 'A': 0, 'Ais': 0, 'H': 0}
    for i in range (0, len(notes)):
        tones[notes[i]] += 1
    
    for x in chords:
        for y in chords[x][0]:
            chords[x][1] += tones[y]
        chords[x][1] /= len(notes)    
  
    sorted_chords = sorted(chords.items(), key = lambda ch:(ch[1][1], ch[0][1]), reverse=True)
    
    chord = []  
    for i in range(0, 3):
        chord.append(sorted_chords[i][0])
    
    return chord
    
'''
    Funkce přiřadí akordům tóny, ze kterých se skládají.
    Vytvoření akordů pro zjišťování pravděpodobností zahraných akordů
    
    @return chords      slovník akordů, každý akord obsahuje seznam tónů, z nichž se skládá
                        a hodnotu 0, do které se bude zaznamenávat počet průniků vyhodnocených tónů
                        se seznamem tónů tvořící akord
'''
def create_chords():
    tones = ['C', 'Cis', 'D', 'Dis', 'E', 'F', 'Fis', 'G', 'Gis', 'A', 'Ais', 'H']
    chords = {}
    
    for i in range (0, len(tones)):
        # dur
        chord = tones[i] + ' dur'
        chords[chord] = [[tones[i%len(tones)], tones[(i+4)%len(tones)], tones[(i+7)%len(tones)], tones[i%len(tones)]], 0]
        # moll
        chord = tones[i] + ' moll'
        chords[chord] = [[tones[i%len(tones)], tones[(i+3)%len(tones)], tones[(i+7)%len(tones)], tones[i%len(tones)]], 0]
        # 7 dur
        chord = tones[i] + '7 dur'
        chords[chord] = [[tones[i%len(tones)], tones[(i+4)%len(tones)], tones[(i+7)%len(tones)], tones[(i+10)%len(tones)]], 0]
        if tones[i] == 'C' or tones[i] == 'D' or tones[i] == 'G':
            # dim
            chord = tones[i] + ' dim'
            chords[chord] = [[tones[i%len(tones)], tones[(i+3)%len(tones)], tones[(i+6)%len(tones)], tones[(i+9)%len(tones)]], 0]
       
    return chords 

'''
    Třída reprezentující vlákno, které na základě spektrální analýzy nahraného zvukového signálu
    vyhodnocuje akordy.
    
    @param QObject  objekt z knihovny PyQt5
'''
class Analysis(QObject):
    # signál ukončující vlákno
    finished = pyqtSignal()
    # signál pro předávání informací o progresu
    progress = pyqtSignal(list)

    '''
        Inicializace počátečních hodnot
        
        @param mic  instance třídy MicrophoneRecorder
    '''
    def __init__(self, mic):
        super(QObject, self).__init__()
        self.mic = mic
        self.frames = []
        self.chords = create_chords()
    
    '''
        Funkce spouštějící a ukončující vlákno
        Dokud probíhá nahrávání, funkce zpracovává zaznamenaný zvukový signál
    '''
    def run(self):
        while(self.mic.isRunning()):
            # nahrané rámce
            frames = self.mic.get_frames()
            
            if len(frames) > 0:
                for frame in frames:
                    # ukladam rámce o poloviční délce
                    self.frames.append(frame[:len(frame)//2])
                    self.frames.append(frame[len(frame)//2:])
                    
            # 3 rámce poloviční délky => 557 ms
            if len(self.frames) >= 3:
                current_frames = self.frames[-3:]            
                self.frames.clear()
                self.analyse(current_frames)
                
        self.finished.emit()    
    
    '''
        Funkce zajišťující analýzu zaznamenaných rámců zvukového signálu
        
        @param frames   rámce, ze kterých budou určeny 3 nejpravděpodobnější akordy
    '''
    def analyse(self, frames):    
        # odhady výkonové spektrální hustoty (PSD), které budou průměrovány
        spect = []
        
        # sloučení polovičních rámců do celých rámců tak, aby se celé rámce překrývaly v polovině 
        windows = []
        for i in range(0, len(frames)-1):
            z = numpy.concatenate((frames[i], frames[i+1]))
            windows.append(z)

        for win in windows:        
            # nahrazení NaN hodnot nulami
            time_series = numpy.nan_to_num(win)

            # je nahrávaný zvuk ticho, nebo hudba ?
            if is_acceptable(time_series) == False: continue
            
            win_array = get_windowing(time_series)        
            fft_array = get_fft(win_array)
            spectrum = get_spectrum(fft_array)
            # Savitzky-Golayův filtr - proložení polynomem řádu 2, k určení křivky použito 23 okolních bodů
            savgol_filt = scipy.signal.savgol_filter(spectrum, 23, 2) 
            
            spect.append(savgol_filt)

        if len(spect) > 0:
            # průměr odhadů PSD
            average = average_spectrals(spect)
            
            # vynásobení odhadu PSD logaritmem
            average_log = get_logarithm(average)

            # stredni hodnota odhadu PSD       
            mean = statistics.mean(average_log)
            # zvyseni o 25 %
            mean *= 1.25 
            # vrcholy odhadu PSD nad střední hodnotou
            peaks = scipy.signal.find_peaks(average_log, height=mean)
            freqs = get_freqs(peaks[0])
            notes = get_notes(freqs)
            
            if len(notes) > 0:
                # vynulování pravděpodobností u jednotlivých akordů
                for x in self.chords:
                    self.chords[x][1] = 0

                chord = get_chord(notes, self.chords)
                # předání výsledku hlavní aplikaci
                self.progress.emit(chord)
