'''
    Bakalářská práce
    Scrolling Song Book - zpěvník s automatickým posunem textu
    
    Hana Kalivodová
    4. 5. 2022
    
    Aplikace zobrazující komponenty, ovládaná uživatelem
'''

import sys
from PyQt5.QtWidgets import *
from PyQt5.QtGui import QTextCharFormat, QColor, QIcon, QFont, QFontMetrics
from PyQt5.QtCore import Qt, QSize, QThread

# soubor zajišťující nahrávání zvuku
import recorder
# soubor zajišťující spektrální analýzu zvuku a rozpoznávání akordů
import analysis
# soubor s konstantami
import const

'''
    Třída zajišťující vykreslení GUI a běh celé aplikace.
    
    @param QMainWindow  hlavní okno knihovny PyQt5
'''
class MainWindow(QMainWindow):

    '''
        Inicializace počátečních hodnot, umístění grafických komponent v aplikaci
        a nastavení reakcí na ně
        
        @param parent   třída, od které tato třída dědí
    '''
    def __init__(self, parent=None):
        super(MainWindow, self).__init__(parent)

        # nastavení okna
        self.setWindowTitle('Scrolling Song Book')
        self.resize(640,480)
        self.move(100, 100)
        self.resizeEvent
        
        # inicializace mikrofonu
        self.mic = recorder.MicrophoneRecorder()
        
        # area pro zobrazení textu
        self.textEdit = QTextEdit()
        self.textEdit.setReadOnly(True)
        self.textEdit.setAlignment(Qt.AlignLeft)
        self.setCentralWidget(self.textEdit)
        
        self.sb = self.textEdit.verticalScrollBar()
        
        # aktualní řádka akordu
        self.row = 0
        # aktualní sloupec akordu
        self.col = 0
        # text písně
        self.text = []
        # akordy písně
        self.chords = []
        # název písně
        self.title = ''    
        # počet odehraných řádek
        self.currentLine = 1
        # posunout text na začátek ?
        self.up = True
        
        # nastavení defaultních hodnot pro zobrazování textu
        self.text_format = QTextCharFormat()
        self.text_format.setFontFamily('Courier')
        self.color = QColor('black')
        self.backColor = QColor('white').name()
        
        toolbar = QToolBar()
        toolbar.setIconSize(QSize(16,16))
        self.addToolBar(toolbar)
        toolbar.setFloatable(False)
        
        # prostor pro zobrazení očekávaného akordu
        self.statusBar = QStatusBar()
        self.setStatusBar(self.statusBar)
        
        # vložení textu s akordy
        btnLoad = QAction('Load lyrics', self)
        btnLoad.triggered.connect(self.loadLyrics)
        toolbar.addAction(btnLoad)
        toolbar.addSeparator()
        
        # font písma
        self.comboFont = QComboBox()
        fonts = ['Courier', 'Consolas', 'Lucida Console']
        self.comboFont.addItems(fonts)
        self.comboFont.activated.connect(self.setFont)
        toolbar.addWidget(self.comboFont)
        
        # velikost písma
        self.spinSize = QSpinBox()
        self.spinSize.setMinimum(10)
        self.spinSize.setMaximum(30)
        self.spinSize.valueChanged.connect(self.setSize)       
        toolbar.addWidget(self.spinSize)
        
        # barva písma
        self.btnColorText = QAction('Color text', self)
        self.btnColorText.triggered.connect(self.setColorText)
        toolbar.addAction(self.btnColorText)
        
        # barva pozadí
        self.btnBackColor = QAction('Background color', self)
        self.btnBackColor.triggered.connect(self.setBackgroundColor)
        toolbar.addAction(self.btnBackColor)                
        toolbar.addSeparator()
        
        label = QLabel(' Transposition ')
        toolbar.addWidget(label)
        
        # možnost transpozice
        self.comboTranspose = QComboBox()
        self.comboTranspose.addItems(const.CHORDS)
        self.comboTranspose.activated.connect(self.transpose)
        toolbar.addWidget(self.comboTranspose)
        toolbar.addSeparator()
        
        # spuštení nahrávání
        self.btnPlay = QAction('Play', self)
        self.btnPlay.triggered.connect(self.startRecording)
        toolbar.addAction(self.btnPlay)
        
        # zastavení nahrávání
        self.btnStop = QAction('Stop', self)
        self.btnStop.triggered.connect(self.stopRecording)
        self.btnStop.setEnabled(False)
        toolbar.addAction(self.btnStop)

        # vlákno pro proces analýzy akordů
        self.thread = QThread()
        
    '''
        Funkce pro přepočet počtu zobrazených řádek v závislosti na změně velikosti okna.
        
        @param  aktuální velikost okna
    '''    
    def resizeEvent(self, newSize):
        self.fontMetrics = QFontMetrics(self.text_format.font())
        self.lines = self.textEdit.viewport().height() // self.fontMetrics.height()
        
    '''
        Funkce pro načtení textu s akordy.
    '''
    def loadLyrics(self):
        self.transposeInit = True
        filename = QFileDialog.getOpenFileName(self, 'Select a File',  '', 'Text files(*.txt*);;all files(*.*)') 
        if filename[0] == '': return
        self.stopRecording()
        self.onlyChords = False
        self.chords.clear()
        self.text.clear()
        self.title = ''
        f = open(filename[0], 'r', encoding='utf-8')
        for line in f:
            self.parseLine(line)
                
        f.close()
                
        self.textEdit.setCurrentCharFormat(self.text_format)
        self.setLyrics() 
        
    '''
        Funkce pro oddělení akordů a textu (1 řádka)
        
        @param line     řádka načteného souboru
    '''
    def parseLine(self, line):   
        if line.startswith('{'):
            # komentar - nazev, zacatek/konec sloky/refrenu, mezihry,...
            if line.startswith('{title:'):
                self.title += line[7:-2].strip()          
            elif line.startswith('{t:'):
                self.title += line[3:-2].strip()
            elif line.startswith('{comment: Instrumental') or line.startswith('{c: Instrumental'):
                self.onlyChords = True
            elif line.startswith('{comment: Solo') or line.startswith('{c: Solo'):
                self.onlyChords = True
            elif self.onlyChords:
                self.onlyChords = False
                self.chords.append('')
                self.text.append('')
            return
            
        if self.onlyChords: 
            # řádka pouze s akordy
            self.parseChords(line)
            return    
            
        size = len(line)
        text = ''
        chords = []
        i = 0
        chord = 0
        
        space = '{:{width}}'.format(' ', width=size) 
        
        while i < size:
            # x = index '[' od i
            x = line[i:].find('[')
            if x == -1:
                text += line[i:]
                break
            if x == 0 and i != 0:
                line = line[:x+i] + ' ' + line[x+i:]
                chords.append(space[:x+1])
                x += 1
            elif x-chord < 0:
                line = line[:x+i] + space[:chord-x+1] + line[x+i:]
                x += chord-x+1
                chords.append(space[:1])
            else:    
                # mezery mezi akordy 
                chords.append(space[:x-chord])
            
            # index od 0    
            x += i
            text += line[i:x]    
            
            # y = index ']' od x 
            y = line[x:].find(']')
    
            # pocet znaku akordu
            chord = y - 1
                
            # index od 0
            y += x
            # pocatek akordu bez zavorek
            x += 1
               
            chords.append(line[x:y]) 
            if self.transposeInit:
                if line[x+1:x+2] == '#' or line[x+1:x+2] == 'b' and line[x:x+1] != 'B':
                    self.comboTranspose.setCurrentText(line[x:x+2])
                else:
                    self.comboTranspose.setCurrentText(line[x:x+1])
                self.transposeInit = False
            
            # text bez zavorek
            i = y + 1
        
        self.chords.append(chords)
        self.text.append(text)
        
    '''
        Funkce pro uložení akordů
        Předpokládá se, že na řádce jsou akordy, které nejsou v [] závorkách.
        
        @param line     řádka akordů
    '''
    def parseChords(self, line):
        if line == '\n': return
        repeat = 1
        splitted = line.split(' ')
        chords = []
        for x in splitted:
            if x.startswith('|') or x == '': continue
            if x.startswith('(x'):
                i = x.find(')')
                repeat = int(x[2:i])
                continue
            chords.append(x)
            chords.append('  ')
        
        for i in range (0, repeat):
            self.chords.append(chords)
            self.text.append('')
        
    '''
        Funkce skládající akordy a text, tak aby akordy příslušné řádky
        byly zobrazeny nad textem dané řádky.
    '''
    def setLyrics(self):
        self.textEdit.setPlainText('')
        self.text_format.setFontWeight(QFont.Bold)
        self.textEdit.setCurrentCharFormat(self.text_format)
        self.textEdit.append(self.title)
        
        for i in range(0, len(self.chords)):
            lyrics = ''
            for x in self.chords[i]:
                self.text_format.setFontWeight(QFont.Bold)
                self.textEdit.setCurrentCharFormat(self.text_format)
                lyrics += x

            self.textEdit.append(lyrics)
            self.text_format.setFontWeight(QFont.Normal)
            self.textEdit.setCurrentCharFormat(self.text_format) 
            self.textEdit.append(self.text[i])

        self.textEdit.append('\n') 
        self.sb.setValue(self.sb.minimum())
        
        self.fontMetrics = QFontMetrics(self.text_format.font())
        # počet zobrazených řádek
        self.lines = self.textEdit.viewport().height() // self.fontMetrics.height()       
        
    '''
        Funkce pro nastavení fontu písma
    '''
    def setFont(self):
        self.text_format.setFontFamily(self.comboFont.currentText())
        self.setLyrics()
        
    '''
        Funkce pro nastavení velikosti písma
    '''
    def setSize(self):
        self.text_format.setFontPointSize(self.spinSize.value())
        self.setLyrics()
    
    '''
        Funkce pro nastavení barvy písma
    '''
    def setColorText(self):
        color = QColorDialog.getColor()
        if not QColor.isValid(color):
            return
        self.color = color
        self.text_format.setForeground(self.color)
        self.setLyrics()
    
    '''
        Funkce pro nastavení pozadí, na kterém je text zobrazen
    '''
    def setBackgroundColor(self):
        color = QColorDialog.getColor()
        if not QColor.isValid(color):
            return
        self.backColor = color.name()
        self.textEdit.setStyleSheet('background:' + self.backColor)
    
    '''
        Funkce zajišťující transpozici akordů
        (zvýšení/snížení všech akordů o stejnou hodnotu - daná uživatelem)
    '''
    def transpose(self):
        chose_i = const.CHORDS.index(self.comboTranspose.currentText())
        chords = []
        firstChord = True
        plus = 0

        for i in range (0, len(self.chords)):
            row = self.chords[i]
            list_row = []
            for j in range (0, len(self.chords[i])):
                chord = row[j]
                if chord.isspace():
                    list_row.append(chord)
                    continue
                if chord == '': continue
                                
                index = const.CHORDS.index(chord[0])
                if firstChord:
                    plus = ((len(const.CHORDS) - index) + chose_i) % len(const.CHORDS)
                    if len(chord) > 1 and (chord[1] == '#' or chord[1] == 'b'):
                        plus = ((len(const.CHORDS) - const.CHORDS.index(chord[:2])) + chose_i) % len(const.CHORDS)
                    firstChord = False
                
                if len(chord) == 1:
                    chord = const.CHORDS[(index+plus) % len(const.CHORDS)]
                    list_row.append(chord)
                    continue
                if chord[1] == '#':
                    index += 1
                    chord = const.CHORDS[(index+plus) % len(const.CHORDS)] + chord[2:]
                    list_row.append(chord)
                    continue
                if chord[1] == 'b':
                    index -= 1
                    if chord[:2] == 'Bb': index += 1
                    chord = const.CHORDS[(index+plus) % len(const.CHORDS)] + chord[2:]
                    list_row.append(chord)
                    continue
                    
                chord = const.CHORDS[(index+plus) % len(const.CHORDS)] + chord[1:]    
                list_row.append(chord)
                
            chords.append(list_row)  
  
        self.chords = chords   
        self.setLyrics()
        
    '''
        Funkce pro spuštění/pozastavení nahrávání zvuku
        Po pozastavení a opětovném spuštění nahrávání uživatel pokračuje v písni, kde skončil.
    '''
    def startRecording(self):
        if self.btnPlay.text() == 'Play':
            if (self.up): 
                self.sb.setValue(self.sb.minimum())
                self.up = False
            self.btnPlay.setText('Pause')
            self.mic.start()
            self.worker = analysis.Analysis(self.mic)
            self.worker.moveToThread(self.thread)
            self.thread.started.connect(self.worker.run)
            self.worker.finished.connect(self.thread.quit)
            self.worker.finished.connect(self.worker.deleteLater)
            self.worker.progress.connect(self.compareChords)
            self.thread.start()
        else:
            self.btnPlay.setText('Play')
            self.mic.stop()
            
        self.btnStop.setEnabled(True)
        
    '''
        Funkce pro zastavení nahrávání
        Po zastavení nahrávání uživatel začíná od začátku písně.
    '''
    def stopRecording(self):
        self.mic.stop()
        self.btnPlay.setEnabled(True)
        self.btnStop.setEnabled(False)
        self.btnPlay.setText('Play')
        self.row = 0
        self.col = 0
        self.currentLine = 1
        self.statusBar.clearMessage()
        self.up = True
            
    '''
        Funkce porovnávající očekávaný akord s vyhodnoceným akordem
        Funkce dává impulz k posunu textu
        
        @param chord    seznam vyhodnocených akordů (výsledek analýzy)
    '''
    def compareChords(self, chord):
        if self.up: return
        if self.chords == []: return
        if self.row >= len(self.chords):
            self.stopRecording()      
            return
        
        currentChord = True
        
        if self.chords[self.row] == []:
            self.row += 1
            self.currentLine += 2
            self.scroll()
        
        if self.row >= len(self.chords):
            self.stopRecording()      
            return

        for i in range(self.col, len(self.chords[self.row])):
            if self.chords[self.row][i].isspace(): continue
            if len(self.chords[self.row][i]) == 0: continue
            if currentChord:
                setChords = self.getSetChords(self.chords[self.row][i])
                # zjisti stejne prvky v polich
                chords = set(setChords).intersection(set(chord))
                if chords != set():
                    currentChord = False
                else:
                    return
            else:
                self.col = i
                return
        
        self.currentLine += 3
        self.scroll()
        self.row += 1
        self.col = 0
        
    '''
        Funkce doplní očekávaný akord o akordy, které jsou s daným akordem akceptované.
        
        @param chord    očekávaný akord z textu písně
        
        @return         seznam akceptovaných akordů pro očekávaný akord z textu
    '''
    def getSetChords(self, chord):
        tones = ['C', 'Cis', 'D', 'Dis', 'E', 'F', 'Fis', 'G', 'Gis', 'A', 'Ais', 'H']

        self.statusBar.showMessage('Next chord: ' + chord)
        
        if chord[0] == 'B':
            chord = 'H' + chord[1:]

        i = tones.index(chord[0])
        
        if len(chord) == 1:
            return [tones[i] + ' dur', tones[i] + '7 dur', tones[(i + 4) % len(tones)] + ' moll']
            
        if chord[1] == 'm':
            return [tones[i] + ' moll', tones[(i + 3) % len(tones)] + ' dur', tones[(i + 3) % len(tones)] + '7 dur']

        elif chord[1] == 'b':
            i -= 1
            if len(chord) > 2 and chord[2] == 'm':
                return [tones[i] + ' moll', tones[(i + 3) % len(tones)] + ' dur', tones[(i + 3) % len(tones)] + '7 dur']
            else:
                return [tones[i] + ' dur', tones[i] + '7 dur', tones[(i + 4) % len(tones)] + ' moll']
        
        elif chord[1] == '#':
            i += 1
            if len(chord) > 2 and chord[2] == 'm':
                return [tones[i] + ' moll', tones[(i + 3) % len(tones)] + ' dur', tones[(i + 3) % len(tones)] + '7 dur']
            else:
                return [tones[i] + ' dur', tones[i] + '7 dur', tones[(i + 4) % len(tones)] + ' moll']
            
        elif chord[1:] == 'dim':
            return [tones[i] + ' dim', tones[(i + 3) % len(tones)] + ' moll', tones[i] + ' dur', tones[i] + '7 dur']
        
        else:
            return [tones[i] + ' dur', tones[i] + '7 dur', tones[(i + 3) % len(tones)] + ' moll'] 
            
    '''
        Funkce zajišťující posun textu
    '''
    def scroll(self):
        # posunutí textu probíhá dříve, než se uživatel dostane na konec stránky (obrazovky)
        if self.currentLine >= (self.lines - 4):
            self.currentLine = 1
            self.sb.setValue(self.sb.value() + (self.lines - 4) * self.fontMetrics.height())

'''
    spuštění aplikace, vytvoření okna            
'''
if __name__ == '__main__':
    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec_())

